"""
Defines the version number of the package.

(-) Release scripts might modify this automatically...
(-) do not add anything but the version number here!

We do it this way, so both __init__.py and setup.py can use it.

see:
https://packaging.python.org/guides/single-sourcing-package-version/
"""

__version__ = "5.0.3"
